#!/usr/bin/env python

"""A set of classes for handling HiC peak calling."""

import os
import sys

import numpy
import h5py

import libraries._hic_peaks as _hic_peaks


class HiCCUPS(object):

    """
    This class calculates the HiCCUPS peak calling described by Rao et al. (2014).

    When initialized, this class creates an h5dict in which to store all data associated with this object.

    :param hic: A :class:`HiC <hifive.hic.HiC>` class object.
    :type hic: :class:`HiC <hifive.hic.HiC>`
    :param binsize: The size in basepairs to bin data into.
    :type binsize: int.
    :param pixel: The width, in bins, of the peaks to be called.
    :type pixel: int.
    :param fdr: The false discovery rate to use for correcting for multiple tests.
    :type fdr: float
    :param chroms: A list of chromosomes to perform peak calling on. If chroms is None, all chroms are used.
    :type chroms: list
    :param silent: Indicates whether to print information about function execution for this object.
    :type silent: bool.

    :attributes: * **file** (*str.*) - A string containing the name of the file passed during object creation for saving the object to.
                 * **silent** (*bool.*) - A boolean indicating whether to suppress all of the output messages.
                 * **history** (*str.*) - A string containing all of the commands executed on this object and their outcome.
    """

    def __init__(self, hic, binsize=25000, pixel=1, fdr=0.05, chroms=None, silent=False):
        """Create a HiCCUPS peak calling object."""
        self.binsize = binsize
        self.minwidth = int(round(15.0 / (binsize ** 0.5)))
        self.maxwidth = 20
        self.pixel = pixel
        self.hic = hic
        self.silent = silent
        self.minobs = 16
        if chroms is None:
            self.chroms = hic.chr2int.keys()
            self.chroms.sort()
        else:
            self.chroms = chroms
        self.scores = {}
        for chrom in self.chroms:
            self.calculate_scores(chrom)

    def calculate_scores(self, chrom):
        """Calculate the HiCCUPS score for a single chromosome.

        :param chrom: The chromosome for which scores are calculated.
        :type chrom: str.
        returns None
        """
        data, mapping = self.hic.cis_heatmap(chrom, binsize=self.binsize, datatype='enrichment', arraytype='upper',
                                             returnmapping=True, silent=self.silent)
        self.scores[chrom] = numpy.zeros(data.shape, dtype=numpy.float32)
        _hic_peaks.calculate_hiccups_scores(
                data,
                self.scores[chrom],
                self.minwidth,
                self.maxwidth,
                self.pixel,
                self.minobs
            )














# distutils: language = c++

"""These functions provide increased speed in handling functions dealing with
3D modeling of HiC data using HiFive.
"""

import cython
cimport numpy as np
import numpy

ctypedef np.float32_t DTYPE_t
ctypedef np.float64_t DTYPE_64_t
ctypedef np.int32_t DTYPE_int_t
ctypedef np.int64_t DTYPE_int64_t
ctypedef np.uint32_t DTYPE_uint_t
ctypedef np.int8_t DTYPE_int8_t
cdef double Inf = numpy.inf

cdef extern from "math.h":
    double exp(double x) nogil
    double log(double x) nogil
    double log10(double x) nogil
    double sqrt(double x) nogil
    double pow(double x, double x) nogil
    double abs(double x) nogil
    double round(double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_distances(
        np.ndarray[DTYPE_t, ndim=2] coords not None,
        np.ndarray[DTYPE_t, ndim=1] distances not None):
    cdef long long int i, j, k, pos
    cdef double value
    cdef long long int num_coords = coords.shape[0]
    cdef long long int num_parameters = coords.shape[1]
    with nogil:
        pos = 0
        for i in range(num_coords - 1):
            for j in range(i + 1, num_coords):
                value = 0.0
                for k in range(num_parameters):
                    value += pow(coords[i, k] - coords[j, k], 2.0)
                distances[pos] = pow(value, 0.5)
                pos += 1
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_poisson_cost(
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices not None,
        np.ndarray[DTYPE_t, ndim=1] zero_weights not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_weights not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_counts not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_log_distances not None,
        np.ndarray[DTYPE_t, ndim=1] distances not None,
        double alpha,
        double beta):
    cdef long long int i
    cdef double cost
    cdef long long int num_zero = zero_indices.shape[0]
    cdef long long int num_nonzero = nonzero_indices.shape[0]
    cdef double log_beta = log(beta)
    with nogil:
        cost = 0.0
        for i in range(num_zero):
            cost -= beta * zero_weights[i] * pow(distances[zero_indices[i]], alpha)
        for i in range(num_nonzero):
            cost += nonzero_counts[i] * (alpha * nonzero_log_distances[i] + log_beta) - beta * nonzero_weights[i] * pow(distances[nonzero_indices[i]], alpha)
    return cost

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_poisson_param_gradients(
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices not None,
        np.ndarray[DTYPE_t, ndim=1] zero_weights not None,
        np.ndarray[DTYPE_t, ndim=1] zero_log_distances not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_weights not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_counts not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_log_distances not None,
        np.ndarray[DTYPE_t, ndim=1] distances not None,
        np.ndarray[DTYPE_64_t, ndim=1] grads not None,
        double alpha,
        double beta):
    cdef long long int i
    cdef double bd_alpha, value
    cdef long long int num_zero = zero_indices.shape[0]
    cdef long long int num_nonzero = nonzero_indices.shape[0]
    with nogil:
        for i in range(num_zero):
            bd_alpha = zero_weights[i] * pow(distances[zero_indices[i]], alpha)
            value = -beta * bd_alpha * zero_log_distances[i]
            grads[0] -= value
            grads[1] += bd_alpha
        for i in range(num_nonzero):
            bd_alpha = nonzero_weights[i] * pow(distances[nonzero_indices[i]], alpha)
            value = (nonzero_counts[i] - beta * bd_alpha) * nonzero_log_distances[i]
            grads[0] += value
            value = nonzero_counts[i] / beta - bd_alpha
            grads[1] += value
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_poisson_coord_gradients(
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] zero_indices1 not None,
        np.ndarray[DTYPE_t, ndim=1] zero_weights not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices0 not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_indices1 not None,
        np.ndarray[DTYPE_t, ndim=1] nonzero_weights not None,
        np.ndarray[DTYPE_int_t, ndim=1] nonzero_counts not None,
        np.ndarray[DTYPE_t, ndim=1] distances not None,
        np.ndarray[DTYPE_t, ndim=2] coords not None,
        np.ndarray[DTYPE_64_t, ndim=2] grads not None,
        double alpha,
        double beta):
    cdef long long int i, j, index0, index1
    cdef double distance, value, value2
    cdef long long int num_zero = zero_indices.shape[0]
    cdef long long int num_nonzero = nonzero_indices.shape[0]
    with nogil:
        for i in range(num_zero):
            distance = distances[zero_indices[i]]
            value = -alpha * beta * zero_weights[i] * pow(distance, alpha) / (distance * distance)
            index0 = zero_indices0[i]
            index1 = zero_indices1[i]
            for j in range(3):
                value2 = value * (coords[index0, j] - coords[index1, j])
                grads[index0, j] -= value2
                grads[index1, j] += value2
        for i in range(num_nonzero):
            distance = distances[nonzero_indices[i]]
            value = alpha * (nonzero_counts[i] - beta * nonzero_weights[i] * pow(distance, alpha)) / (distance * distance)
            index0 = nonzero_indices0[i]
            index1 = nonzero_indices1[i]
            for j in range(3):
                value2 = value * (coords[index0, j] - coords[index1, j])
                grads[index0, j] -= value2
                grads[index1, j] += value2
    return None













#!/usr/bin/env python

import sys

try:
    from mpi4py import MPI
except:
    pass

from ..fend import Fend
from ..hic_data import HiCData
from ..hic import HiC


def run(args):
    if 'mpi4py' in sys.modules.keys():
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        num_procs = comm.Get_size()
    else:
        comm = None
        rank = 0
        num_procs = 1
    if args.chroms is None:
        chroms = []
    else:
        chroms = args.chroms.split(',')
        if len(chroms) == 1 and chroms[0] == '':
            chroms = []
    if args.matrix is not None and args.binned is None:
        if rank == 0:
            print sys.stderr, ("Loading data from matrices is only supported for binned data.\n")
        return 1
    if args.algorithm.count('binning') > 0:
        if args.binned is not None:
            if rank == 0:
                print sys.stderr, ("This normalization algorithm is not currently supported for binned data.\n")
            return 1
        model = args.model.split(',')
        modelbins = args.modelbins.split(',')
        parameters = args.parameters.split(',')
        for i in range(len(modelbins)):
            try:
                modelbins[i] = int(modelbins[i])
            except:
                if rank == 0:
                    print sys.stderr, ("Not all arguments in -n/--modelbins could be converted to integers.\n")
                return 1
        if len(model) != len(modelbins) or len(model) != len(parameters):
            if rank == 0:
                print sys.stderr, ("-v/--model, -n/--modelbins, and -u/--parameter-types must be equal lengths.\n")
            return 1
    if args.binned == 0 and args.bed is None:
        if rank == 0:
            print  >> sys.stderr, ("Non-uniforming binning (binned=0) must have a bed file to read bin partitions from.\n"),
        return None
    elif args.binned is not None and args.binned < 1 and args.length is not None:
        if rank == 0:
            print  >> sys.stderr, ("Binning from a chromosome length file needs a positive integer value for binning.\n"),
        return None
    if args.prefix is None:
        fend_fname, data_fname, project_fname = args.output
    else:
        fend_fname = "%s.fends" % args.prefix
        data_fname = "%s.hcd" % args.prefix
        project_fname = "%s.hcp" % args.prefix
    if rank == 0:
        fends = Fend(fend_fname, mode='w', binned=args.binned, silent=args.silent)
        if args.bed is not None:
            if args.binned is not None and args.binned == 0:
                fends.load_bins(args.bed, genome_name=args.genome, format='bed')
            else:
                fends.load_fends(args.bed, genome_name=args.genome, re_name=args.re, format="bed")
        elif args.fend is not None:
            fends.load_fends(args.fend, genome_name=args.genome, re_name=args.re, format="fend")
        else:
            fends.load_bins(args.length, genome_name=args.genome, format='len')
        fends.save()
        del fends
        data = HiCData(data_fname, 'w', silent=args.silent)
        if not args.bam is None: 
            data.load_data_from_bam(fend_fname, args.bam, args.insert, args.skipdups)
        elif not args.raw is None: 
            data.load_data_from_raw(fend_fname, args.raw, args.insert, args.skipdups)
        elif not args.mat is None: 
            data.load_data_from_mat(fend_fname, args.mat, args.insert)
        elif not args.matrix is None:
            data.load_binned_data_from_matrices(fend_fname, args.matrix, format=None)
        data.save()
        del data
        for i in range(1, num_procs):
            comm.send(1, dest=i, tag=11)
    else:
        comm.recv(source=0, tag=11)
    hic = HiC(project_fname, 'w', silent=args.silent)
    hic.load_data(data_fname)
    hic.filter_fends(mininteractions=args.minint, mindistance=args.mindist, maxdistance=args.maxdist)
    hic.find_distance_parameters(minsize=args.minbin, numbins=args.numbins)
    precorrect = False
    if args.algorithm in ['binning', 'binning-express', 'binning-probability']:
        hic.find_binning_fend_corrections(mindistance=args.mindist, maxdistance=args.maxdist, parameters=parameters,
                                             chroms=chroms, num_bins=modelbins, model=model, usereads=args.binreads,
                                             learning_threshold=args.threshold, max_iterations=args.biniter,
                                             pseudocounts=args.pseudo)
        precorrect = True
    if args.algorithm in ['probability', 'binning-probability']:
        hic.find_probability_fend_corrections(mindistance=args.mindist, maxdistance=args.maxdist,
                                              minchange=args.change, max_iterations=args.probiter,
                                              learningstep=args.step, chroms=chroms,
                                              precalculate=args.precalc, precorrect=precorrect)
    elif args.algorithm in ['express', 'binning-express']:
        hic.find_express_fend_corrections(iterations=args.expiter, mindistance=args.mindist,
                                          maxdistance=args.maxdist, remove_distance=args.nodist,
                                          usereads=args.expreads, mininteractions=args.minint,
                                          chroms=chroms, minchange=args.change, precorrect=precorrect,
                                          binary=args.binary, kr=args.kr)
    if rank == 0:
        hic.save()
